/* Copyright (C) 2001/2002 The Honeynet Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by The Honeynet Project.
 * 4. The name "The Honeynet Project" may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#pragma ident "@(#)$Id: sebek.c,v 1.14 2003/11/30 18:37:58 job Exp $"

#include "sebek.h"

/*
 * Module Loading and Unloading
 * See modctl.h for external structure definitions.
 * The type misc does not get any other hooks in the kernel.
 * The type strmod is used to prepare for loading the filtering module.
 */

/* STREAMS module declaration structure */
static struct fmodsw fsw = {
  "sbk",
  &sbkinfo,
  D_NEW | D_MP
};

/*
 * Module linkage information for the kernel
 */

/* Module description */
static struct modlstrmod modlstrmod = {
  &mod_strmodops,
  "sbk",
  &fsw
};

static struct modlinkage modlinkage = {
    MODREV_1, {(void *) &modlstrmod, NULL}
};

char  _depends_on[] = "drv/ip";

#ifdef  SEBEK_DEBUG
int sebek_debug = 1;
#else
int sebek_debug = 0;
#endif                          /* SEBEK_DEBUG */

int loaded = 0;

/*
 * Structure of the system-entry table.
 */
#ifdef BITS64
extern struct sysent sysent32[];
#endif
extern struct sysent sysent[];

static void unlink_mod(struct modctl *mp);
static void link_mod(struct modctl *mp);
static void print_mod(struct modctl *mp);
static void list_ifs();
static void if_attach(queue_t *in, queue_t *out);
static void if_detach();
static void do_autopush();
static void clear_autopush();
static int open_stream(struct vnode **vpp);
static void close_stream(struct vnode *vp);

int64_t (*old_read) (int fd, char *buf, size_t count);
int64_t (*old_read32) (int fd, char *buf, size_t count);
int (*old_spec_ioctl)(struct vnode *vp, int cmd, intptr_t arg, int mode, struct cred *cr, int *rvalp);

kmutex_t sebek_lock;
struct qinit *qinitp = NULL;
static struct modctl *my_mp = NULL;
int stream_open = 0;
struct vnodeops *spec_vnodeopsp;
major_t network_devmaj = 0;

int _init(void)
{
    struct modctl *mp;
    int s;
    ill_t *il = NULL;
    dev_t device;
    char buf[128];

    if (loaded) {
        sebek_log(SEBEK_CE_DEBUG1, "sebek initializing while loaded!\n");
        return(EINVAL);
    }

    sebek_log(SEBEK_CE_DEBUG1, "sebek initializing\n");

    if ((il = ill_lookup_on_name(INTERFACE, strlen(INTERFACE)+1, FALSE, FALSE)) == NULL) {
        sebek_log(SEBEK_CE_DEBUG1, "could not find interface with ipv4!\n");
        return(EINVAL);
    }

#if defined(SEBEK_HIDE)
    /*
     * We go and hunt our module. The module chain is a loop so we use the
     * head to detect the end. When found we change the name so in case any
     * unlinking fails, the original will be properly loaded later.
     */

    for (mp = modules.mod_next; mp != &modules; mp = mp->mod_next) {
        if (strncmp("sebekhide", mp->mod_modname,9) == 0) {
            sebek_log(SEBEK_CE_DEBUG1, "found myself!\n");
            my_mp = mp;
            print_mod(my_mp);
            break;
        }
    }

    if (my_mp) {
        sebek_log(SEBEK_CE_DEBUG1, "unlinking module\n");
        unlink_mod(my_mp);
    }
#endif

    mutex_init(&sebek_lock, NULL, MUTEX_DEFAULT, NULL);

    spec_vnodeopsp = spec_getvnodeops();

    /*
     * Configure ethernet destination addr.
     * (make source identical for convenience)
     */

    ether_aton(DESTINATION_MAC, &ether_dst);
    ether_aton(DESTINATION_MAC, &ether_src);

    /*
     * Took the info for local IP from ip_filter.
     */

#ifdef ILLF_IPV4
    if (il && (il->ill_flags&ILLF_IPV4)==ILLF_IPV4) {
        ip_local.s_addr = il->ill_ipif->ipif_local_addr;
        sebek_log(SEBEK_CE_DEBUG1, "Has IPV4 address %lx",ip_local.s_addr);
    } else ip_local.s_addr = inet_addr(DESTINATION_IP);
#else
    if (il && il->ill_ipif &&
        (il->ill_ipif->ipif_flags&IFF_IPV4)==IFF_IPV4) {
        ip_local.s_addr = il->ill_ipif->ipif_local_addr;
        sebek_log(SEBEK_CE_DEBUG1, "Has IPV4 address %lx",ip_local.s_addr);
    } else ip_local.s_addr = inet_addr(DESTINATION_IP);
#endif

    ip_dst.s_addr = inet_addr(DESTINATION_IP);

    /*
     * open the stream before autopushing our own module.
     * it is on the same stream.
     */

    if (open_stream(&sbkvp)) {
        sebek_log(SEBEK_CE_DEBUG1, "Opening stream to nw device failed\n");
        return(EINVAL);
    } else stream_open = 1;

    sebek_log(SEBEK_CE_DEBUG1, "stream opened\n");

    /*
     * Save the old values.
     */

    old_read = (int64_t (*)()) sysent[SYS_read].sy_callc;
#ifdef BITS64
    old_read32 = (int64_t (*)()) sysent32[SYS_read].sy_callc;
#endif
    old_spec_ioctl = spec_vnodeopsp->vop_ioctl;

    /*
     * Switch to critical to reduce the risk of races and swap the
     * function pointers.
     */

    s = ddi_enter_critical();

    sysent[SYS_read].sy_callc = (int64_t(*)())new_read;
#ifdef BITS64
    sysent32[SYS_read].sy_callc = (int64_t(*)())new_read32;
#endif
    spec_vnodeopsp->vop_ioctl = new_spec_ioctl;

/* postponed till later. Not sure if our packets can also
 * come up the /dev/ip stream.

    if (il && il->ill_rq && il->ill_wq && il->ill_wq->q_next) {
        if_attach(il->ill_rq,il->ill_wq->q_next);
    }

*/

    ddi_exit_critical(s);

    /*
     * Some actions need to be delayed. In this case configuration of autopush.
     * At this moment we are not known yet in the module list.
     */

    timeout(delayed_init, NULL, drv_usectohz(1000000));

    loaded = 1;

    /*
     * We assume mod_install will *not* fail. This is actually not
     * true and may result in system corruption. Recovery to be implemented.
     */

    return (mod_install(&modlinkage));
}

int _fini()
{
    int s;
    struct backd *backdp;

    /* This may happen when the stream is busy */
    if (!loaded) {
        sebek_log(SEBEK_CE_DEBUG1, "sebek _fini while not loaded!\n");
        return (mod_remove(&modlinkage));
    }

    if (sbkvp) {
        stream_open = 0;
        close_stream(sbkvp);
        sbkvp = NULL;
    }
        
    /*
     * Switch to critical to reduce the risk of races and swap the
     * function pointers.
     */

    s = ddi_enter_critical();

    sysent[SYS_read].sy_callc = (int64_t(*)())old_read;
#ifdef BITS64
    sysent32[SYS_read].sy_callc = (int64_t(*)())old_read32;
#endif
    spec_vnodeopsp->vop_ioctl = old_spec_ioctl;

/*
    if (old_rqinit && old_wqinit) {
        if_detach();
    }
*/

    ddi_exit_critical(s);

    /* remove all the module structs */
    mutex_enter(&sebek_lock);

    /* remove some stuff, nohthing here yet */

    mutex_exit(&sebek_lock);

    /* no idea if this is the way to safely destroy a mutex */
    while (!mutex_tryenter(&sebek_lock));
    mutex_destroy(&sebek_lock);

    clear_autopush();

    sebek_log(SEBEK_CE_DEBUG1, "sebek removed\n");

    loaded = 0;

    return (mod_remove(&modlinkage));
}

int _info(struct modinfo * modinfop)
{   
    struct modctl *mp;
    int i;

    sebek_log(SEBEK_CE_DEBUG1, "sebek info called\n");

    return (mod_info(&modlinkage, modinfop));
}

/*
 * delayed init can be used to delay until some required modules have been
 * loaded. Particularly the autopush feature for the filtering module.
 */
void delayed_init()
{
    sebek_log(SEBEK_CE_DEBUG1, "delayed_init called\n");

    do_autopush();
}

static void unlink_mod(struct modctl *mp)
{   
    struct modctl *prevmp, *nextmp;

    mutex_enter(&mod_lock);

    nextmp = mp->mod_next;
    prevmp = mp->mod_prev;
    prevmp->mod_next = nextmp;
    nextmp->mod_prev = prevmp;
    last_module_id--;

    mutex_exit(&mod_lock);
}

static void link_mod(struct modctl *mp)
{
    mutex_enter(&mod_lock);
 
    mp->mod_next = &modules;
    mp->mod_prev = modules.mod_prev;
    modules.mod_prev->mod_next = mp;
    modules.mod_prev = mp;

    mutex_exit(&mod_lock);
}

void sebek_log(int level, const char *fmt, ...)
{
    auto char buf[1024];
    va_list ap;

    va_start(ap, fmt);
    (void) vsprintf(buf, fmt, ap);
    va_end(ap);

    switch (level) {
    case CE_CONT:
    case CE_NOTE:
    case CE_WARN:
    case CE_PANIC:
        cmn_err(level, "sebek:\t%s", buf);
        break;

    case SEBEK_CE_DEBUG4:
        if (sebek_debug < 4)
            break;
     /*FALLTHROUGH*/ case SEBEK_CE_DEBUG3:
        if (sebek_debug < 3)
            break;
     /*FALLTHROUGH*/ case SEBEK_CE_DEBUG2:
        if (sebek_debug < 2)
            break;
     /*FALLTHROUGH*/ case SEBEK_CE_DEBUG1:
        if (sebek_debug < 1)
            break;
        cmn_err(CE_CONT, "sebek:\t%s", buf);
     /*FALLTHROUGH*/ default:
        break;
    }
}

static void print_mod(struct modctl *mp)
{   

    sebek_log(SEBEK_CE_DEBUG1, "mod_id = %d\n", mp->mod_id);
    sebek_log(SEBEK_CE_DEBUG1, "mod_mp = %lx\n", mp->mod_mp);
    sebek_log(SEBEK_CE_DEBUG1, "mod_filename = %s\n", mp->mod_filename);
    sebek_log(SEBEK_CE_DEBUG1, "mod_modname = %s\n", mp->mod_modname);
    sebek_log(SEBEK_CE_DEBUG1, "mod_busy = %d\n", mp->mod_busy);
    sebek_log(SEBEK_CE_DEBUG1, "mod_stub = %d\n", mp->mod_stub);
    sebek_log(SEBEK_CE_DEBUG1, "mod_loaded = %x\n", mp->mod_loaded);
    sebek_log(SEBEK_CE_DEBUG1, "mod_installed = %x\n", mp->mod_installed);
    sebek_log(SEBEK_CE_DEBUG1, "mod_loadflags = %x\n", mp->mod_loadflags);
    sebek_log(SEBEK_CE_DEBUG1, "mod_want = %x\n", mp->mod_want);
    sebek_log(SEBEK_CE_DEBUG1, "mod_loadcnt = %d\n", mp->mod_loadcnt);

}

static void list_ifs()
{
    queue_t *in, *out;
    ill_t *il;
#if SOL_MINVER >= 10
    ill_walk_context_t ctx;
#endif

    sebek_log(SEBEK_CE_DEBUG1, "list_ifs()\n");

#if SOL_MINVER >= 10
    for (il = ILL_START_WALK_ALL(&ctx); il; il = ill_next(&ctx, il)) {
#else   
    for (il = ill_g_head; il; il = il->ill_next) {
#endif  

        in = il->ill_rq;
        if (!in || !il->ill_wq)
            continue;
        out = il->ill_wq->q_next;

        sebek_log(SEBEK_CE_DEBUG1, "ill_name %s\n",il->ill_name);
        sebek_log(SEBEK_CE_DEBUG1, "ill_ndd_name: %s\n",il->ill_ndd_name);
        if (il->ill_rq && il->ill_rq->q_qinfo && il->ill_rq->q_qinfo->qi_minfo) {
             sebek_log(SEBEK_CE_DEBUG1, "idname rq: %s\n", il->ill_rq->q_qinfo->qi_minfo->mi_idname);
        }
        if (il->ill_wq && il->ill_wq->q_qinfo && il->ill_wq->q_qinfo->qi_minfo) {
             sebek_log(SEBEK_CE_DEBUG1, "idname wq: %s\n", il->ill_wq->q_qinfo->qi_minfo->mi_idname);
        }
        if (il->ill_wq->q_next && il->ill_wq->q_next->q_qinfo && il->ill_wq->q_next->q_qinfo->qi_minfo) {
             sebek_log(SEBEK_CE_DEBUG1, "idname wq next: %s\n", il->ill_wq->q_next->q_qinfo->qi_minfo->mi_idname);
        }
        sebek_log(SEBEK_CE_DEBUG1, "sap %lx\n",il->ill_sap);
        sebek_log(SEBEK_CE_DEBUG1, "type %x\n",il->ill_type);

    }

}

static void if_attach(queue_t *in, queue_t *out)
{
    if (in && out) {
        qinp = in;
        qoutp = out;

        old_rqinit = in->q_qinfo;
        old_wqinit = out->q_qinfo;

        bcopy((caddr_t)old_wqinit, (caddr_t)&new_wqinit,
              sizeof(struct qinit));
        new_wqinit.qi_putp = ipwput;
        out->q_qinfo = &new_wqinit;

        bcopy((caddr_t)old_rqinit, (caddr_t)&new_rqinit,
              sizeof(struct qinit));
        new_rqinit.qi_putp = iprput;
        in->q_qinfo = &new_rqinit;
    }
}

static void if_detach()
{
    if (qinp && qoutp) {
        qinp->q_qinfo = old_rqinit;
        qoutp->q_qinfo = old_wqinit;
    }
}

/*
 * This does not only open the stream, but also sends 1 packet as a test.
 * Heavy in development!
 */
static int open_stream(struct vnode **vpp)
{
    int n = 0;
    queue_t *wq;
    struct vnode *vp = NULL;
    int err;
    dl_attach_req_t *dl_attach_reqp;
    dl_bind_req_t *dl_bind_reqp;
    dl_info_req_t *dl_info_reqp;
    dl_unitdata_req_t *dl_unitdata_reqp;
    char *cp;
    int rval;
    mblk_t *mp;
    struct strioctl sioc;
    struct ether_header *ehp;
    rval_t retval;
    char modifname[MODMAXNAMELEN+1];
    char ifn;
    dev_t device;

    strcpy(modifname,INTERFACE);
    ifn = modifname[strlen(modifname)-1];
    modifname[strlen(modifname)-1] = '\0';

    /*
     * the network device is a cloned device. So we construct it with
     * the major number 11 of clone and minor numer equal to network major
     */

    if ((network_devmaj = ddi_name_to_major(modifname)) < 0) {
        sebek_log(SEBEK_CE_DEBUG1, "unknown major\n");
        return (-1);
    }

    sebek_log(SEBEK_CE_DEBUG1, "major device %s:%d\n",INTERFACE, network_devmaj);
    device = makedevice(11,network_devmaj);

    sebek_log(SEBEK_CE_DEBUG1, "openening stream for device %d\n", getmajor(device));

    if (device == 0 || device == NODEV) {
        sebek_log(SEBEK_CE_DEBUG1, "wrong device\n");
        return (-1);
    }

    *vpp = vp;
    vp = makespecvp(device, VCHR);

    do {
        if ((err = VOP_OPEN(&vp, (FREAD|FWRITE|FNDELAY), kcred)) == EAGAIN) {
            (void) delay(hz);
        }
    } while (err == EAGAIN && ++n < 5);

    if (vp->v_stream == NULL) {
        sebek_log(SEBEK_CE_DEBUG1, "Unexpected: no stream\n");
        VOP_CLOSE(vp, 0, 1, (offset_t)0, kcred);
        VN_RELE(vp);
        return (-1);
    }

    wq = vp->v_stream->sd_wrq;
    while(wq) {
        if (wq->q_qinfo && wq->q_qinfo->qi_minfo) {
            sebek_log(SEBEK_CE_DEBUG1, "Module %s\n",wq->q_qinfo->qi_minfo->mi_idname);
        }
        wq = wq->q_next;
    }

    *vpp = vp;

    sebek_log(SEBEK_CE_DEBUG1, "opened stream\n");
   
    /*
     * Attaching to PPA. This is the instance of the device.
     */ 
    mp = makemblk(sizeof (dl_attach_req_t), M_PROTO);
    dl_attach_reqp = (dl_attach_req_t *)(mp->b_rptr);
    dl_attach_reqp->dl_primitive = DL_ATTACH_REQ;
    dl_attach_reqp->dl_ppa = ifn - '0';

    err = send_and_ack(vp, &mp, &retval);

    if (err) {
        sebek_log(SEBEK_CE_DEBUG1, "attach error %d\n",err);
        VOP_CLOSE(vp, 0, 1, (offset_t)0, kcred);
        VN_RELE(vp);
        return(-1);
    }

    if (((union DL_primitives *)(mp->b_rptr))->dl_primitive != DL_OK_ACK) {
        sebek_log(SEBEK_CE_DEBUG1, "Unexpected attach return prim = %d\n",
            ((union DL_primitives *)(mp->b_rptr))->dl_primitive);
    }
    freemsg(mp);

    /*
     * Binding to a SAP.
     * In this case 0 for default.
     */ 
    mp = makemblk(sizeof (dl_bind_req_t), M_PROTO);
    dl_bind_reqp = (dl_bind_req_t *)(mp->b_rptr);
    dl_bind_reqp->dl_primitive = DL_BIND_REQ;
    dl_bind_reqp->dl_sap = 0;
    dl_bind_reqp->dl_max_conind = 0;
    dl_bind_reqp->dl_service_mode = DL_CLDLS;
    dl_bind_reqp->dl_conn_mgmt = 0;
    dl_bind_reqp->dl_xidtest_flg = 0;

    err = send_and_ack(vp, &mp, &retval);

    if (err) {
        sebek_log(SEBEK_CE_DEBUG1, "bind error %d\n",err);
        VOP_CLOSE(vp, 0, 1, (offset_t)0, kcred);
        VN_RELE(vp);
        return(-1);
    }

    if (((union DL_primitives *)(mp->b_rptr))->dl_primitive != DL_BIND_ACK) {
        sebek_log(SEBEK_CE_DEBUG1, "Unexpected bind return prim = %d\n",
            ((union DL_primitives *)(mp->b_rptr))->dl_primitive);
    }
    freemsg(mp);


    /*
     * Request info from the device.
     * This gives info on physical address size etc. Can be used
     * to distinguish device types.
     */ 
    mp = makemblk(sizeof (dl_info_req_t), M_PROTO);
    dl_info_reqp = (dl_info_req_t *)(mp->b_rptr);
    dl_info_reqp->dl_primitive = DL_INFO_REQ;

    err = send_and_ack(vp, &mp, &retval);

    if (err) {
        sebek_log(SEBEK_CE_DEBUG1, "info_req error %d\n",err);
        VOP_CLOSE(vp, 0, 1, (offset_t)0, kcred);
        VN_RELE(vp);
        return(-1);
    }

    if (((union DL_primitives *)(mp->b_rptr))->dl_primitive != DL_INFO_ACK) {
        sebek_log(SEBEK_CE_DEBUG1, "Unexpected info return prim = %d\n",
            ((union DL_primitives *)(mp->b_rptr))->dl_primitive);
    } else {
        dl_info_ack_t * dl_info_ackp = (dl_info_ack_t *)(mp->b_rptr);

        sebek_log(SEBEK_CE_DEBUG1,"saplen %d, total %d\n",myABS(dl_info_ackp->dl_sap_length),
            dl_info_ackp->dl_addr_length);
    }
    freemsg(mp);

    /*
     * Set the device to raw operation.
     */ 
    sioc.ic_cmd = DLIOCRAW;
    sioc.ic_timout = 0;
    sioc.ic_len = 0;
    sioc.ic_dp = 0;

    err = strioctl(vp, I_STR, (intptr_t)&sioc, 0, K_TO_K, kcred, &rval);
    if (err) {
        sebek_log(SEBEK_CE_DEBUG1, "ack error %d\n",err);
        VOP_CLOSE(vp, 0, 1, (offset_t)0, kcred);
        VN_RELE(vp);
        return(-1);
    }
 
    sebek_log(SEBEK_CE_DEBUG1, "ioctl retval %d\n",rval);


    return(0);
}

static void close_stream(struct vnode *vp)
{
    VOP_CLOSE(vp, 0, 1, (offset_t)0, kcred);
    VN_RELE(vp);
}

/*
 * Autopush our filter on the device that is used for sending of
 * the log packets. After successful opening of the stream the major
 * of the device should be present in network_devmaj.
 */

static void do_autopush()
{
    struct strapush ap;
    struct vnode *vp;
    dev_t device;
    int sadmaj;
    int rval;
    int err;

    ap.sap_cmd = SAP_ALL;
    ap.sap_major = network_devmaj;
    ap.sap_minor = 0;
    ap.sap_lastminor = 0;
    ap.sap_npush = 1;
    strcpy(ap.sap_list[0], "sbk");

    sadmaj = ddi_name_to_major("sad");
    device = makedevice(sadmaj, ADMMIN);
    vp = makespecvp(device, VCHR);

    if ((err = VOP_OPEN(&vp, (FREAD|FWRITE), kcred)) != 0) {
        sebek_log(SEBEK_CE_DEBUG1, "autopush: opening device failed\n");
        return;
    }

    if (vp->v_stream == NULL) {
        sebek_log(SEBEK_CE_DEBUG1, "Unexpected: no stream\n");
        VOP_CLOSE(vp, 0, 1, (offset_t)0, kcred);
        VN_RELE(vp);
        return;
    }

    if ((err = strioctl(vp, SAD_SAP, (intptr_t)&ap, 0, K_TO_K, kcred, &rval)) != 0) {
        sebek_log(SEBEK_CE_DEBUG1, "autopush: ioctl failed\n");
        return;
    }

    VOP_CLOSE(vp, 0, 1, (offset_t)0, kcred);
    VN_RELE(vp);
    sebek_log(SEBEK_CE_DEBUG1, "autopush: success\n");
}

static void clear_autopush()
{
    struct strapush ap;
    dev_t device;
    struct vnode *vp;
    int sadmaj;
    int rval;
    int err;
    
    ap.sap_cmd = SAP_CLEAR;
    ap.sap_major = network_devmaj;
    ap.sap_minor = 0;
    ap.sap_lastminor = 0;
    ap.sap_npush = 0;

    sadmaj = ddi_name_to_major("sad");
    device = makedevice(sadmaj, ADMMIN);
    vp = makespecvp(device, VCHR);

    if ((err = VOP_OPEN(&vp, (FREAD|FWRITE), kcred)) != 0) {
        sebek_log(SEBEK_CE_DEBUG1, "autopush: opening device failed\n");
        return;
    }

    if (vp->v_stream == NULL) {
        sebek_log(SEBEK_CE_DEBUG1, "Unexpected: no stream\n");
        VOP_CLOSE(vp, 0, 1, (offset_t)0, kcred);
        VN_RELE(vp);
        return;
    }

    if ((err = strioctl(vp, SAD_SAP, (intptr_t)&ap, 0, K_TO_K, kcred, &rval)) != 0) {
        sebek_log(SEBEK_CE_DEBUG1, "autopush: ioctl failed\n");
        return;
    }

    VOP_CLOSE(vp, 0, 1, (offset_t)0, kcred);
    VN_RELE(vp);
    sebek_log(SEBEK_CE_DEBUG1, "clear autopush: success\n");
}

/*
 * Compare memory regions.
 */
int
memcmp(s1, s2, n)
        const void *s1, *s2;
        size_t n;
{
        if (n != 0) {
                register const unsigned char *p1 = s1, *p2 = s2;

                do {
                        if (*p1++ != *p2++)
                                return (*--p1 - *--p2);
                } while (--n != 0);
        }
        return (0);
}

